/**
    Copyright (c) 2021 NIBIO <http://www.nibio.no/>. 

    This file is part of Model_GENPHENOLO.  
    Model_GENPHENOLO is free software: you can redistribute it and/or modify  
    it under the terms of the NIBIO Open Source License as published by  
    NIBIO, either version 1 of the License, or (at your option) any 
    later version. 

    Model_GENPHENOLO is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    NIBIO Open Source License for more details. 

    You should have received a copy of the NIBIO Open Source License 
    along with Model_GENPHENOLO.  If not, see <http://www.nibio.no/licenses/>.
 */

package no.nibio.vips.model.phenology;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.TimeZone;
import no.nibio.vips.entity.ModelConfiguration;
import no.nibio.vips.entity.Result;
import no.nibio.vips.entity.ResultImpl;
import no.nibio.vips.entity.WeatherObservation;
import no.nibio.vips.i18n.I18nImpl;
import no.nibio.vips.model.ConfigValidationException;
import no.nibio.vips.model.Model;
import no.nibio.vips.model.ModelExcecutionException;
import no.nibio.vips.model.ModelId;
import no.nibio.vips.util.CommonNamespaces;
import no.nibio.vips.util.ModelUtil;
import no.nibio.vips.util.WeatherUtil;

/**
 *
 * @author Bhabesh Bhabani Mukhopadyay
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */
public class PhenologyModel extends I18nImpl implements Model{

    public final static String              NAME_MODEL_ID       =   "GENPHENOLO";
    public final static ModelId             MODEL_ID            =   new ModelId(NAME_MODEL_ID);
    
    private   final     ModelUtil           modelUtil;
    private             DataMatrix          dataMatrix;
    
    private             TimeZone            timeZone;
    private             List<Plantation>    plantations;
    
    private             int                 algorithmId;
    private             Double              baseTemp;
    private             Double              optTemp;  
    private             Double              upperTemp; 
   
    
    public PhenologyModel(String resourceBundleBaseName) 
    {
        super(resourceBundleBaseName);
        this.modelUtil    = new ModelUtil();
    }

   public PhenologyModel() 
   {
       super("no.nibio.vips.model.phenology.texts");
       this.modelUtil    = new ModelUtil();
   }
   
   /**
    * 
    * @return
    * @throws ModelExcecutionException 
    */
    @Override
    public List<Result> getResult() throws ModelExcecutionException {
        
        PhenologyModelStore modelStore  = PhenologyModelStore.getInstance();
        
        List<Result>    results         =   new ArrayList<>();
        Date            currentDate     =   this.dataMatrix.getFirstDateWithParameterValue(DataMatrix.TEMPERATURE_MEAN);
        Date            endDate         =   this.dataMatrix.getLastDateWithParameterValue(DataMatrix.TEMPERATURE_MEAN);
        
        DecimalFormat   dFormat         =   new DecimalFormat("###.##");
        Calendar        cal             =   Calendar.getInstance(timeZone);

        Double          heatSum         =   0.0;
        Double          GDD             =   0.0;
        
        while(currentDate.before(endDate))
        {
            Result              result              =   new ResultImpl();
            
            
            Double              tm_daily            =   Double.parseDouble(dFormat.format(this.dataMatrix.getParamValueForDate(currentDate, DataMatrix.TEMPERATURE_MEAN)));
            
                                switch(algorithmId)
                                {
                                    case 1:
                                        heatSum             =   modelStore.dayDegreeModel(tm_daily, heatSum, baseTemp);
                                        break;
                                    case 2 :    
                                         heatSum         =   heatSum + Double.parseDouble(dFormat.format(this.dataMatrix.getParamValueForDate(currentDate, DataMatrix.TEMPERATURE_DAILY_TIME)));
                                         //System.out.println( currentDate + ": " + " DTT : " + this.dataMatrix.getParamValueForDate(currentDate, DataMatrix.TEMPERATURE_DAILY_TIME) +  " GDD : "+heatSum);
                                         break;
                                         
                                }
            
            
            String              mazeVariety         =   this.dataMatrix.getParamStringValueForDate(currentDate, DataMatrix.PARAM_PHENOLOGYM_VARIETY);
            

            
            
            //if(tm_nonNegative > 0)
            {    
            result.setValidTimeStart(currentDate);
                for (Plantation plantation: plantations)
                {
                        if(mazeVariety.equals(plantation.getTypeName()))
                        {
                            if(heatSum >= plantation.getHeatReq())
                            {
                                result.setValue(NAME_MODEL_ID, DataMatrix.PARAM_PHENOLOGYM_PHASE, plantation.getPhaseName());
                                result.setValue(NAME_MODEL_ID, DataMatrix.PARAM_HEAT_SUM, dFormat.format(heatSum));
                                result.setValue(NAME_MODEL_ID, DataMatrix.PARAM_HEAT_REQUIREMENT, dFormat.format(plantation.getHeatReq()));
                                
                                break;
                            }
                        }
                 }
                
                for (Plantation plantation: plantations)
                {
                    result.setValue("THRESHOLD", plantation.getPhaseName().toUpperCase(), dFormat.format(plantation.getHeatReq()));
                }
                
                
            result.setValue(CommonNamespaces.NS_WEATHER, DataMatrix.TEMPERATURE_MEAN, dFormat.format(this.dataMatrix.getParamValueForDate(currentDate, DataMatrix.TEMPERATURE_MEAN)));
            result.setValue(CommonNamespaces.NS_WEATHER, DataMatrix.TEMP_NON_NEGATIVE, dFormat.format(modelStore.getNonNegative(tm_daily, baseTemp)));
            result.setValue(NAME_MODEL_ID, DataMatrix.PARAM_PHENOLOGYM_VARIETY, mazeVariety);
            results.add(result);
            }
            cal.setTime(currentDate);
            cal.add(Calendar.DATE, 1);
            
            
            
            currentDate = cal.getTime(); 
            
        }
        Collections.sort(results);
        //System.out.println("dataMatrix : "+this.dataMatrix.toString());
        return results;
    }

    /**
     * 
     * @return 
     */
    @Override
    public ModelId getModelId() {
        return PhenologyModel.MODEL_ID;
    }

    @Override
    public String getModelName() {
        return this.getModelName(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelName(String language) {
        
        return this.getText("name", language);
    }

    @Override
    public String getLicense() {
        return this.getText("license", Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getCopyright() {
        return "(c) 2021 NIBIO (http://www.nibio.no/). Contact: post@nibio.no";
    }

    @Override
    public String getModelDescription() {
        return this.getModelDescription(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelDescription(String language) {
        try
        {
            return this.modelUtil.getTextWithBase64EncodedImages(this.getText("description", language), this.getClass());
        }
        catch(IOException ex)
        {
            return this.getText("description", language);
        }
    }

    @Override
    public String getWarningStatusInterpretation() {
        return this.getWarningStatusInterpretation(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getWarningStatusInterpretation(String language) {
        return this.getText("statusInterpretation", language);
    }

    @Override
    public String getModelUsage() {
        return this.getModelUsage(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelUsage(String language) {
         return this.getText("usage", language);
    }

    @Override
    public String getSampleConfig() {
        return this.getText("sampleconfig", Model.DEFAULT_LANGUAGE);
    }

    @Override
    public void setConfiguration(ModelConfiguration config) throws ConfigValidationException {
        
                                    dataMatrix                                      =   new DataMatrix();
        ObjectMapper                mapper                                          =   new ObjectMapper();        
        List<WeatherObservation>    observations                                    =   mapper.convertValue
                                                                                        (
                                                                                                config.getConfigParameter("observations")
                                                                                            ,   new TypeReference<List<WeatherObservation>>(){}
                                                                                        );
        List<WeatherObservation>    fawWeatherLIstHourly_tm                         =   new ArrayList<>();
        Map <Date, WeatherObservation>  fawWeatherMapHourly_tx                      =   new HashMap<> ();
        Map <Date, WeatherObservation>  fawWeatherMapHourly_tn                      =   new HashMap<> ();
        Map <Date, WeatherObservation>  fawWeatherMapDaily_tx                       =   new HashMap<> ();
        Map <Date, WeatherObservation>  fawWeatherMapDaily_tn                       =   new HashMap<> ();
        
        PhenologyModelStore             modelStore                                  =   PhenologyModelStore.getInstance();
        
        String                          flagTM_TX_TN                                =   null;
        WeatherUtil wUtil = new WeatherUtil();
        Date                        dateHourlyTm_currentDay                         =   null; 
        Date                        dateHourlyTm_previousDay                        =   null; 
        Date                        dateSowTrimmed                                  =   null;
        
        Plant                       plantationData                                  =   (null == mapper.convertValue(config.getConfigParameter(DataMatrix.PLANTATION_DATA), new TypeReference<Plant>(){}))
                                                                                        ? null 
                                                                                        : mapper.convertValue(config.getConfigParameter(DataMatrix.PLANTATION_DATA), new TypeReference<Plant>(){});
                                      
                                    plantations                                     =   plantationData.getPlantations();
                                    
                                    algorithmId                                     =   plantationData.getAlgorithmId();
                                    baseTemp                                        =   plantationData.getBaseTemp();
                                    optTemp                                         =   plantationData.getOptTemp();
                                    upperTemp                                       =   plantationData.getUpperTemp();
       
        this.timeZone = TimeZone.getTimeZone((String) config.getConfigParameter("timeZone"));
        Collections.sort(observations);
        dateSowTrimmed  = trimmedDate(wUtil.pragmaticAdjustmentToMidnight(plantationData.getStartDate(),timeZone),timeZone);
        
        for(WeatherObservation weatherObj: observations)
        {
            
            Date woTime = weatherObj.getTimeMeasured();
            //System.out.println(" weatherObj.getTimeMeasured() : "+weatherObj.getTimeMeasured());
            
            Double      HTT             = 0.0;                                  /* Hourly Tharmal Time */
            Double      TM              = 0.0;                                  /* Mean Temperature */
            
            
            weatherObj.setTimeMeasured(wUtil.pragmaticAdjustmentToMidnight(woTime, timeZone));

                     dateHourlyTm_currentDay                                     =   trimmedDate(weatherObj.getTimeMeasured(),timeZone);
                     
                     
                        if(weatherObj.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
                        {
                            WeatherObservation      fawWeatherBO_tm_hourly    =   weatherObj;
                            if  (
                                        (null != dateHourlyTm_currentDay )
                                    &&  (null != dateHourlyTm_previousDay)
                                    &&  dateHourlyTm_currentDay.after   (
                                                                            dateHourlyTm_previousDay
                                                                        )
                                )
                            {
                                
                                if(dateHourlyTm_currentDay.after(dateSowTrimmed))
                                {
                                    
                                    
                                    double  counterTMHourly     =    0;
                                    Double  DTT_add             =   0.0;        /* Daily thermal time  */
                                    Double  DTT                 =   0.0;        /* Daily thermal time  */
                                    Double  HTT_posetive_count  =   0.0;

                                    if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MEAN))
                                    {
                                                flagTM_TX_TN    = DataMatrix.TEMPERATURE_MEAN;
                                                
                                                //TODO calculation of accumulation value and add to daily
                                                for(WeatherObservation wo: fawWeatherLIstHourly_tm)
                                                {
                                                                TM                  =   wo.getValue();
                                                                counterTMHourly     =   counterTMHourly + wo.getValue();
                                                                
                                                                HTT                 =   modelStore.hourlyMeanWithBaseMaxAndOptTemp(TM,baseTemp,optTemp, upperTemp);
                                                                DTT_add             =   HTT + DTT_add;
                                                                if(HTT > 0)
                                                                {
                                                                    HTT_posetive_count = HTT_posetive_count + 1;
                                                                }
                                                                //System.out.println("Data 1 : "+woTime +" : " +"TM :"+TM+ " baseTemp: "+baseTemp+" optTemp:"+optTemp+" upperTemp: "+upperTemp+" HTT : "+HTT+" DTT_add : "+DTT_add);
                                                                

                                                }
                                    }
                                    else
                                    {
                                        flagTM_TX_TN    =   (weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MAX))?DataMatrix.TEMPERATURE_MAX:DataMatrix.TEMPERATURE_MIN;
                                        
                                        if(! fawWeatherMapHourly_tx.isEmpty() && ! fawWeatherMapHourly_tn.isEmpty())
                                        {
                                           for(Map.Entry<Date, WeatherObservation> entry : fawWeatherMapHourly_tx.entrySet())
                                           {
                                               if(null != fawWeatherMapHourly_tn.get(entry.getKey()))
                                                       {
                                                           WeatherObservation tx          = entry.getValue();
                                                           WeatherObservation tn          = fawWeatherMapHourly_tn.get(entry.getKey());
                                                                        TM                = (tx.getValue() + tn.getValue())/2;
                                                                         counterTMHourly  =  counterTMHourly + TM;
                                                                        HTT               =   modelStore.hourlyMeanWithBaseMaxAndOptTemp(TM,baseTemp,optTemp, upperTemp);
                                                                        DTT_add           =   HTT + DTT_add;   
                                                                        
                                                                        if(HTT > 0)
                                                                        {
                                                                            HTT_posetive_count = HTT_posetive_count + 1;
                                                                        }
                                                                         
                                                       }
                                           }
                                        }
                                    }
                                     
                                        /* Mostly supports algorithm id 2 */
                                            
                                           
                                    
                                            
                                    int size = 0;
                                    
                                    //average value of temperature for a day
                                    if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MEAN)) 
                                    {
                                        size    =   fawWeatherLIstHourly_tm.size();
                                        DTT     = DTT_add / size;
                                       /*
                                        if(HTT_posetive_count > 0)
                                        {
                                            DTT     = DTT_add / HTT_posetive_count;
                                        }
                                        */
                                        dataMatrix.setParamDoubleValueForDate   (dateHourlyTm_previousDay
                                                                                ,   DataMatrix.TEMPERATURE_MEAN
                                                                                ,   (
                                                                                        counterTMHourly
                                                                                        /
                                                                                        size
                                                                                    )
                                                                                );
                                    
                                    }
                                    else // For TN/TX
                                    {
                                        
                                        size    = (fawWeatherMapHourly_tx.size() <= fawWeatherMapHourly_tn.size()) ? fawWeatherMapHourly_tx.size() : fawWeatherMapHourly_tn.size();
                                      DTT     = DTT_add / size;
                                        /*
                                        if(HTT_posetive_count > 0)
                                        {
                                            DTT     = DTT_add / HTT_posetive_count;
                                        }
                                        */
                                        dataMatrix.setParamDoubleValueForDate   (dateHourlyTm_previousDay
                                                                                ,   DataMatrix.TEMPERATURE_MEAN
                                                                                ,   (
                                                                                        counterTMHourly
                                                                                        /
                                                                                        size
                                                                                    )
                                                                            );

                                        
                                    }
                                    
                                    if(algorithmId==2) 
                                        {
                                                                                
                                            
                                                                                dataMatrix.setParamDoubleValueForDate   (
                                                                                                                            dateHourlyTm_previousDay
                                                                                                                        ,   DataMatrix.TEMPERATURE_DAILY_TIME
                                                                                                                        ,   DTT
                                                                                                                        );
                                        }

                                    dataMatrix.setParamStringValueForDate(dateHourlyTm_previousDay, DataMatrix.PARAM_PHENOLOGYM_VARIETY, plantationData.getTypeName());
                                
                                }
                                
                                    //discard old value for next day and start fresh
                                    fawWeatherLIstHourly_tm  =   new ArrayList<WeatherObservation>();
                                    fawWeatherMapHourly_tx   =   new HashMap<Date, WeatherObservation> ();
                                    fawWeatherMapHourly_tn   =   new HashMap<Date, WeatherObservation> ();
                            }

                                    fawWeatherLIstHourly_tm.add(fawWeatherBO_tm_hourly);
                                    
                                    if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MAX))
                                     {
                                         fawWeatherMapHourly_tx.put(dateHourlyTm_currentDay, fawWeatherBO_tm_hourly);
                                     }
                                     
                                     if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MIN))
                                     {
                                         fawWeatherMapHourly_tn.put(dateHourlyTm_currentDay, fawWeatherBO_tm_hourly);
                                     } 
                                    
                        }
                        
                        if(weatherObj.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                        {
                            if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MAX))
                            {
                                fawWeatherMapDaily_tx.put(weatherObj.getTimeMeasured(), weatherObj);
                            }
                            else if(weatherObj.getElementMeasurementTypeId().equals(DataMatrix.TEMPERATURE_MIN))
                            {
                                fawWeatherMapDaily_tn.put(weatherObj.getTimeMeasured(), weatherObj);
                            }
                            if(! fawWeatherMapDaily_tx.isEmpty())
                            {
                                Date    txTnDate = null;
                                double  tmValue  = 0.0;
                                for(Map.Entry<Date, WeatherObservation> entry : fawWeatherMapDaily_tx.entrySet())
                                {
                                    if(null != fawWeatherMapDaily_tn.get(entry.getKey()))
                                    {
                                        txTnDate    =   entry.getKey();
                                        tmValue     =   (fawWeatherMapDaily_tn.get(entry.getKey()).getValue() + fawWeatherMapDaily_tx.get(entry.getKey()).getValue()) / 2;
                                        break;
                                    }    
                                        
                                }
                                
                                if (null != txTnDate)
                                {
                                    fawWeatherMapDaily_tx.remove(txTnDate);
                                    fawWeatherMapDaily_tn.remove(txTnDate);
                                    
                                                                    
                                    dataMatrix.setParamDoubleValueForDate   (
                                                                                   txTnDate  //TODO - Might be the date need to be trimmed
                                                                                ,   DataMatrix.TEMPERATURE_MEAN
                                                                                ,   tmValue
                                                                            );

                                    dataMatrix.setParamStringValueForDate(txTnDate , DataMatrix.PARAM_PHENOLOGYM_VARIETY, plantationData.getTypeName());
                                    
                                    txTnDate = null;
                                    tmValue  = 0.0;
                                
                                }
                            
                                
                            
                                    if  (
                                                (null == dataMatrix.getParamDoubleValueForDate(dateHourlyTm_currentDay, DataMatrix.TEMPERATURE_MEAN) )
                                            || 
                                                ((dataMatrix.getParamDoubleValueForDate(dateHourlyTm_currentDay, DataMatrix.TEMPERATURE_MEAN)).intValue() == 0)
                                        )
                                        {
                                            if(dateHourlyTm_currentDay.after(dateSowTrimmed)){
                                                                                                dataMatrix.setParamDoubleValueForDate   (
                                                                                                                                                weatherObj.getTimeMeasured()  //TODO - Might be the date need to be trimmed
                                                                                                                                            ,   DataMatrix.TEMPERATURE_MEAN
                                                                                                                                            ,   weatherObj.getValue()
                                                                                                                                        );

                                                                                                dataMatrix.setParamStringValueForDate(weatherObj.getTimeMeasured() , DataMatrix.PARAM_PHENOLOGYM_VARIETY, plantationData.getTypeName());
                                                                                                }

                                        }
                        }
         
            }
                        dateHourlyTm_previousDay = dateHourlyTm_currentDay;
        }
   
    }

    
    
   /**
     * Get a trimmed date without hour,minute,second,millisecond
     * @param date
     * @return 
     */
    private Date trimmedDate(Date date, TimeZone timezone)
    {
        Date        resultDate  = date; 
        Calendar     calendar   =  Calendar.getInstance();
        calendar.setTimeZone(timezone);
        calendar.setTime(resultDate);
        calendar.set(Calendar.HOUR_OF_DAY, 0);
        calendar.set(Calendar.MINUTE,0);
        calendar.set(Calendar.SECOND,0);
        calendar.set(Calendar.MILLISECOND,0);
        
        return   calendar.getTime();
    }    
    


  
    
    
}
